# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2023 Benjamin Thomas Schwertfeger
# All rights reserved.
# https://github.com/btschwertfeger
#

"""Module that implements the setup functions of the trading algorithm."""

from __future__ import annotations

import traceback
from logging import getLogger
from typing import TYPE_CHECKING, Self

if TYPE_CHECKING:
    from kraken_infinity_grid.gridbot import KrakenInfinityGridBot

LOG = getLogger(__name__)


class SetupManager:
    """SetupManager class to manage the setup of the trading algorithm."""

    def __init__(self: SetupManager, strategy: KrakenInfinityGridBot) -> None:
        LOG.debug("Initializing SetupManager...")
        self.__s = strategy

    def __update_orderbook_get_open_orders(self: SetupManager) -> tuple[list, list]:
        """Get the open orders and txid as lists."""
        LOG.info("  - Retrieving open orders from upstream...")

        open_orders, open_txids = [], []
        for txid, order in self.__s.user.get_open_orders(
            userref=self.__s.userref,
        )["open"].items():
            if order["descr"]["pair"] == self.__s.altname:
                order["txid"] = txid  # IMPORTANT
                open_orders.append(order)
                open_txids.append(order["txid"])
        return open_orders, open_txids

    def __update_order_book_handle_closed_order(
        self: SetupManager,
        closed_order: dict,
    ) -> None:
        """
        Gets executed when an order of the local orderbook was closed in the
        upstream orderbook during the ``update_orderbook`` function in the init
        of the algorithm.

        This function triggers the Telegram message of the executed order and
        places a new order.
        """
        LOG.info("Handling executed order: %s", closed_order["txid"])
        closed_order["side"] = closed_order["descr"]["type"]

        message = str(
            f"✅ {self.__s.symbol}: {closed_order['side'][0].upper()}{closed_order['side'][1:]} "
            "order executed"
            f"\n ├ Price » {closed_order['price']} {self.__s.quote_currency}"
            f"\n ├ Size » {closed_order['vol_exec']} {self.__s.base_currency}"
            f"\n └ Size in {self.__s.quote_currency} » "
            f"{float(closed_order['price']) * float(closed_order['vol_exec'])}",
        )

        self.__s.t.send_to_telegram(message)

        # ======================================================================
        # If a buy order was filled, the sell order needs to be placed.
        if closed_order["side"] == "buy":
            self.__s.om.handle_arbitrage(
                side="sell",
                order_price=self.__s.get_order_price(
                    side="sell",
                    last_price=float(closed_order["price"]),
                ),
                txid_to_delete=closed_order["txid"],
            )

        # ======================================================================
        # If a sell order was filled, we may need to place a new buy order.
        elif closed_order["side"] == "sell":
            # A new buy order will only be placed if there is another sell
            # order, because if the last sell order was filled, the price is so
            # high, that all buy orders will be canceled anyway and new buy
            # orders will be placed in ``check_price_range`` during shift-up.
            if (
                self.__s.orderbook.count(
                    filters={"side": "sell"},
                    exclude={"txid": closed_order["txid"]},
                )
                != 0
            ):
                self.__s.om.handle_arbitrage(
                    side="buy",
                    order_price=self.__s.get_order_price(
                        side="buy",
                        last_price=float(closed_order["price"]),
                    ),
                    txid_to_delete=closed_order["txid"],
                )
            else:
                self.__s.orderbook.remove(filters={"txid": closed_order["txid"]})

    def __update_order_book(self: SetupManager) -> None:
        """
        This function only gets triggered once during the setup of the
        algorithm.

        It checks:
        - ... if the orderbook is up to date, remove filled, closed, and
          canceled orders.
        - ... the local orderbook for changes - comparison with upstream
          orderbook
        - ... and will place new orders if filled.
        """
        LOG.info("- Syncing the orderbook with upstream...")

        # ======================================================================
        # Only track orders that belong to this instance.
        ##
        open_orders, open_txids = self.__update_orderbook_get_open_orders()

        # ======================================================================
        # Orders of the upstream which are not yet tracked in the local
        # orderbook will now be added to the local orderbook.
        ##
        local_txids = [order["txid"] for order in self.__s.orderbook.get_orders()]
        something_changed = False
        for order in open_orders:
            if order["txid"] not in local_txids:
                LOG.info(
                    "  - Adding upstream order to local orderbook: %s",
                    order["txid"],
                )
                self.__s.orderbook.add(order)
                something_changed = True
        if not something_changed:
            LOG.info("  - Nothing changed!")

        # ======================================================================
        # Check all orders of the local orderbook against those from upstream.
        # If they got filled -> place new orders.
        # If canceled -> remove from local orderbook.
        ##
        for order in self.__s.orderbook.get_orders():
            if order["txid"] not in open_txids:
                closed_order = self.__s.om.get_orders_info_with_retry(
                    txid=order["txid"],
                )
                # ==============================================================
                # Order was filled
                if closed_order["status"] == "closed":
                    self.__update_order_book_handle_closed_order(
                        closed_order=closed_order,
                    )

                # ==============================================================
                # Order was closed
                elif closed_order["status"] in {"canceled", "expired"}:
                    self.__s.orderbook.remove(filters={"txid": order["txid"]})

                else:
                    # pending || open order - still active
                    ##
                    continue

        # There are no more filled/closed and cancelled orders in the local
        # orderbook and all upstream orders are tracked locally.
        LOG.info("- Orderbook initialized!")

    def __check_asset_pair_parameter(self: Self) -> None:
        """Check the asset pair parameter."""
        LOG.info("- Checking asset pair parameters...")
        pair_data = self.__s.market.get_asset_pairs(
            pair=[self.__s.symbol.replace("/", "")],
        )
        LOG.debug(pair_data)

        self.__s.xsymbol = next(iter(pair_data.keys()))
        data = pair_data[self.__s.xsymbol]

        self.__s.altname = data["altname"]
        self.__s.zbase_currency = data["base"]  # XXBT
        self.__s.xquote_currency = data["quote"]  # ZEUR
        self.__s.cost_decimals = data["cost_decimals"]  # 5, i.e., 0.00001 EUR

        if self.__s.fee is None:
            # This is the case if the '--fee' parameter was not passed, then we
            # take the highest maker fee.
            self.__s.fee = float(data["fees_maker"][0][1]) / 100

        self.__s.amount_per_grid_plus_fee = self.__s.amount_per_grid * (
            1 + self.__s.fee
        )

    def __check_configuration_changes(self: Self) -> None:
        """
        Checking if the database content match with the setup parameters.

        Checking if the order size or the interval have changed, requiring
        all open buy orders to be cancelled.
        """
        LOG.info("- Checking configuration changes...")
        cancel_all_orders = False

        if self.__s.amount_per_grid != self.__s.configuration.get()["amount_per_grid"]:
            LOG.info(" - Amount per grid changed => cancel open buy orders soon...")
            self.__s.configuration.update({"amount_per_grid": self.__s.amount_per_grid})
            cancel_all_orders = True

        if self.__s.interval != self.__s.configuration.get()["interval"]:
            LOG.info(" - Interval changed => cancel open buy orders soon...")
            self.__s.configuration.update({"interval": self.__s.interval})
            cancel_all_orders = True

        if cancel_all_orders:
            self.__s.om.cancel_all_open_buy_orders()

        LOG.info("- Configuration checked and up-to-date!")

    def prepare_for_trading(self: Self) -> None:
        """
        This function gets triggered once during the setup of the algorithm. It
        prepares the algorithm for live trading by checking the asset pair
        parameters, syncing the local with the upstream orderbook, place missing
        sell orders that not get through because of e.g. "missing funds", and
        updating the orderbook.
        """
        LOG.info(
            "Preparing for trading by initializing and updating local orderbook...",
        )

        self.__s.t.send_to_telegram(
            message=f"✅ {self.__s.name} - {self.__s.symbol} is live again!",
            exception=True,
            log=False,
        )
        # ======================================================================

        # Check the fee and altname of the asset pair
        ##
        self.__check_asset_pair_parameter()

        # Append orders to local orderbook in case they are not saved yet
        ##
        self.__s.om.assign_all_pending_transactions()

        # Try to place missing sell orders that not get through because
        # of "missing funds".
        ##
        self.__s.om.add_missed_sell_orders()

        # Update the orderbook, check for closed, filled, cancelled trades,
        # and submit new orders if necessary.
        ##

        try:
            self.__update_order_book()
        except Exception as e:  # noqa: BLE001
            message = f"Exception in update_orderbook: {e}: {traceback.format_exc()}"
            self.__s.save_exit(message)

        # Check if the configured amount per grid or the interval have changed,
        # requiring a cancellation of all open buy orders.
        ##
        self.__check_configuration_changes()

        # Everything is done, the bot is ready to trade live.
        ##
        self.__s.is_ready_to_trade = True
        LOG.info("Algorithm is ready to trade!")

        # Checks if the open orders match the range and cancel if
        # necessary. It is the heart of this algorithm and gets
        # triggered every time the price changes.
        ##
        self.__s.om.check_price_range()
        self.__s.init_done = True



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2024 Benjamin Thomas Schwertfeger
# All rights reserved.
# https://github.com/btschwertfeger
#

"""Module implementing the database connection and handling of interactions."""

from datetime import datetime
from importlib.metadata import version
from logging import getLogger
from typing import Any, Self

from sqlalchemy import (
    Column,
    DateTime,
    Float,
    Integer,
    MetaData,
    String,
    Table,
    asc,
    create_engine,
    delete,
    desc,
    func,
    select,
    text,
    update,
)
from sqlalchemy.engine.result import MappingResult
from sqlalchemy.orm import sessionmaker

LOG = getLogger(__name__)


class DBConnect:
    """Class handling the connection to the PostgreSQL or sqlite database."""

    def __init__(  # pylint: disable=too-many-positional-arguments
        self: Self,
        db_user: str | None = None,
        db_password: str | None = None,
        db_host: str | None = None,
        db_port: str | int | None = None,
        db_name: str = "kraken_infinity_grid",
        in_memory: bool = False,
        sqlite_file: str | None = None,
    ) -> None:
        LOG.info("Connecting to the database...")
        if in_memory:
            engine = "sqlite:///:memory:"
        elif sqlite_file:
            engine = f"sqlite:///{sqlite_file}"
        else:
            engine = "postgresql://"
            if db_user and db_password:
                engine += f"{db_user}:{db_password}@"
            if db_host and db_port:
                engine += f"{db_host}:{db_port}"
            engine += f"/{db_name}"

        self.engine = create_engine(engine)
        self.session = sessionmaker(bind=self.engine)()
        self.metadata = MetaData()

    def init_db(self: Self) -> None:
        """Create tables if they do not exist and pre-fill with default rows."""
        LOG.info("- Initializing tables...")
        self.metadata.create_all(self.engine)
        LOG.info("- Database initialized.")

    def add_row(self: Self, table: Table, **kwargs: Any) -> None:
        """Insert a row into the specified table."""
        LOG.debug("Inserting a row into '%s': %s", table, kwargs)
        self.session.execute(table.insert().values(**kwargs))
        self.session.commit()

    def get_rows(
        self: Self,
        table: Table,
        filters: dict | None = None,
        exclude: dict | None = None,
        order_by: tuple[str, str] | None = None,  # (column_name, "asc" or "desc")
        limit: int | None = None,
    ) -> MappingResult:
        """Fetch rows from the specified table with optional filters, ordering, and limit."""
        LOG.debug(
            "Querying rows from table '%s' with filters: %s, order_by: %s, limit: %s",
            table,
            filters,
            order_by,
            limit,
        )
        query = select(table)
        if filters:
            query = query.where(
                *(table.c[column] == value for column, value in filters.items()),
            )
        if exclude:
            query = query.where(
                *(table.c[column] != value for column, value in exclude.items()),
            )
        if order_by:
            column, direction = order_by
            if direction.lower() == "asc":
                query = query.order_by(asc(table.c[column]))
            elif direction.lower() == "desc":
                query = query.order_by(desc(table.c[column]))
        if limit:
            query = query.limit(limit)
        return self.session.execute(query).mappings()

    def update_row(
        self: Self,
        table: Table,
        filters: dict,
        updates: dict,
    ) -> None:
        """Update rows in the specified table matching filters."""
        LOG.debug("Update rows from '%s': %s :: %s", table, filters, updates)
        query = (
            update(table)
            .where(*(table.c[column] == value for column, value in filters.items()))
            .values(**updates)
        )
        self.session.execute(query)
        self.session.commit()

    def delete_row(self: Self, table: Table, filters: dict) -> None:
        """Delete rows from the specified table matching filters."""
        LOG.debug("Deleting row(s) from '%s': %s", table, filters)
        query = delete(table).where(
            *(table.c[column] == value for column, value in filters.items()),
        )
        self.session.execute(query)
        self.session.commit()


class Orderbook:
    """Table containing the orderbook data."""

    def __init__(self: Self, userref: int, db: DBConnect) -> None:
        LOG.debug("Initializing the orderbook table...")
        self.__db = db
        self.__userref = userref
        self.__table = Table(
            "orderbook",
            db.metadata,
            Column("id", Integer, primary_key=True),
            Column("userref", Integer, nullable=False),
            Column("txid", String, nullable=False),
            Column("symbol", String, nullable=False),
            Column("side", String, nullable=False),
            Column("price", Float, nullable=False),
            Column("volume", Float, nullable=False),
        )

    def add(self: Self, order: dict) -> None:
        """Add an order to the orderbook."""
        LOG.debug("Adding order to the orderbook: %s", order)
        self.__db.add_row(
            self.__table,
            userref=self.__userref,
            txid=order["txid"],
            symbol=order["descr"]["pair"],
            side=order["descr"]["type"],
            price=order["descr"]["price"],
            volume=order["vol"],
        )

    def get_orders(
        self: Self,
        filters: dict | None = None,
        exclude: dict | None = None,
        order_by: tuple[str, str] | None = None,
        limit: int | None = None,
    ) -> MappingResult:
        """Get orders from the orderbook."""
        LOG.debug(
            "Getting orders from the orderbook with filters: %s, exclude: %s, order_by: %s, limit: %s",
            filters,
            exclude,
            order_by,
            limit,
        )
        if not filters:
            filters = {}
        return self.__db.get_rows(
            self.__table,
            filters=filters | {"userref": self.__userref},
            exclude=exclude,
            order_by=order_by,
            limit=limit,
        )

    def remove(self: Self, filters: dict) -> None:
        """Remove orders from the orderbook."""
        LOG.debug("Removing orders from the orderbook: %s", filters)
        if not filters:
            raise ValueError("Filters required for removal in orderbook")
        self.__db.delete_row(
            self.__table,
            filters=filters | {"userref": self.__userref},
        )

    def update(self: Self, updates: dict, filters: dict | None = None) -> None:
        """
        Update orders in the orderbook.

        In case one manually modifies the order. This is not recommended!
        """
        LOG.debug("Updating orders in the orderbook: %s :: %s", filters, updates)
        if not filters:
            filters = {}

        prepared_updates = {}
        if "txid" in updates:
            prepared_updates["txid"] = updates["txid"]

        if descr := updates.get("descr"):
            if descr.get("pair"):
                prepared_updates["symbol"] = descr["pair"]
            if descr.get("type"):
                prepared_updates["side"] = descr["type"]  # should not happen
            if descr.get("price"):
                prepared_updates["price"] = descr["price"]

        if "vol" in updates:
            prepared_updates["volume"] = updates["vol"]

        self.__db.update_row(
            self.__table,
            filters=filters | {"userref": self.__userref},
            updates=prepared_updates,
        )

    def count(
        self: Self,
        filters: dict | None = None,
        exclude: dict | None = None,
    ) -> int:
        """Count orders in the orderbook."""
        LOG.debug(
            "Counting orders in the orderbook with filters: %s and exclude: %s",
            filters,
            exclude,
        )
        if not filters:
            filters = {}
        filters |= {"userref": self.__userref}

        query = (
            select(func.count())  # pylint: disable=not-callable
            .select_from(self.__table)
            .where(
                *(self.__table.c[column] == value for column, value in filters.items()),
            )
        )
        if exclude:
            query = query.where(
                *(self.__table.c[column] != value for column, value in exclude.items()),
            )
        return self.__db.session.execute(query).scalar()  # type: ignore[no-any-return]


class Configuration:
    """Table containing information about the bots config."""

    def __init__(self: Self, userref: int, db: DBConnect) -> None:
        LOG.debug("Initializing the configuration table...")
        self.__db = db
        self.__userref = userref
        self.__table = Table(
            "configuration",
            self.__db.metadata,
            Column("id", Integer, primary_key=True),
            Column("userref", Integer, nullable=False),
            Column(
                "version",
                String,
                nullable=False,
                default=version("kraken-infinity-grid"),
            ),
            Column("vol_of_unfilled_remaining", Float, nullable=False, default=0),
            Column(
                "vol_of_unfilled_remaining_max_price",
                Float,
                nullable=False,
                default=0,
            ),
            Column("price_of_highest_buy", Float, nullable=False, default=0),
            Column("amount_per_grid", Float),
            Column("interval", Float),
            Column("last_price_time", DateTime, nullable=False),
            Column("last_telegram_update", DateTime, nullable=False),
            extend_existing=True,
        )

        # Create if not exist
        self.__table.create(bind=self.__db.engine, checkfirst=True)

        # self.__migrate_table()

        # Add initial values
        if not self.__db.get_rows(
            self.__table,
            filters={"userref": self.__userref},
        ).fetchone():  # type: ignore[no-untyped-call]
            self.__db.add_row(
                self.__table,
                userref=self.__userref,
                last_price_time=datetime.now(),
                last_telegram_update=datetime.now(),
            )

    def __migrate_table(self: Self) -> None:
        """
        Migrate the table to the latest version.

        This method may be extended to handle more complex migrations in the
        future, for now its just a placeholder.
        """
        LOG.info("- Check for %s table migration...", self.__table.name)
        any_change = False

        # ======================================================================
        # Renaming columns
        ##
        change_column = text(
            "SELECT column_name FROM information_schema.columns"
            " WHERE table_name = :table_name"
            " AND column_name = :column_name",
        )
        if not self.__db.session.execute(
            change_column,
            {"table_name": self.__table.name, "column_name": "version"},
        ).fetchone():
            LOG.info(
                "  - Adding column 'version' to %s table...",
                self.__table.name,
            )
            self.__db.session.execute(
                text(
                    f"ALTER TABLE {self.__table.name} ADD COLUMN version VARCHAR",
                ),
            )
            any_change = True

        if not self.__db.session.execute(
            change_column,
            {"table_name": self.__table.name, "column_name": "amount_per_grid"},
        ).fetchone():
            LOG.info(
                "  - Adding column 'amount_per_grid' to %s table...",
                self.__table.name,
            )
            self.__db.session.execute(
                text(
                    f"ALTER TABLE {self.__table.name}"
                    " ADD COLUMN amount_per_grid FLOAT",
                ),
            )
            any_change = True

        if not self.__db.session.execute(
            change_column,
            {"table_name": self.__table.name, "column_name": "interval"},
        ).fetchone():
            LOG.info("  - Adding column 'interval' to %s table...", self.__table.name)
            self.__db.session.execute(
                text(f"ALTER TABLE {self.__table.name} ADD COLUMN interval FLOAT"),
            )
            any_change = True

        self.__db.session.execute(
            text(
                f"UPDATE {self.__table.name}"  # noqa: S608
                " SET version = :version WHERE userref = :userref",
            ),
            {"version": version("kraken-infinity-grid"), "userref": self.__userref},
        )

        if any_change:
            LOG.info("  - Migration of %s table successfully!", self.__table.name)

    def get(self: Self, filters: dict | None = None) -> dict:
        """Get configuration from the table."""
        LOG.debug(
            "Getting configuration from the table 'configuration' with filter: %s",
            filters,
        )
        if not filters:
            filters = {}

        if result := self.__db.get_rows(
            self.__table,
            filters=filters | {"userref": self.__userref},
        ):
            return next(result)
        raise ValueError(f"No configuration found for passed {filters=}!")

    def update(self: Self, updates: dict) -> None:
        """Update configuration in the table."""
        LOG.debug("Updating configuration in the table: %s", updates)
        self.__db.update_row(
            self.__table,
            filters={"userref": self.__userref},
            updates=updates,
        )


class UnsoldBuyOrderTXIDs:
    """
    Table containing information about future sell orders. Entries are added
    before placing a new sell order in order to not miss the placement in case
    placing fails.

    If the placement succeeds, the entry gets deleted from this table.
    """

    def __init__(self: Self, userref: int, db: DBConnect) -> None:
        LOG.debug("Initializing the UnsoldBuyOrderTXIDs table...")
        self.__db = db
        self.__userref = userref
        self.__table = Table(
            "unsold_buy_order_txids",
            self.__db.metadata,
            Column("id", Integer, primary_key=True),
            Column("userref", Integer, nullable=False),
            Column("txid", String, nullable=False),  # corresponding buy order
            Column("price", Float, nullable=False),  # price at which to sell
        )

    def add(self: Self, txid: str, price: float) -> None:
        """Add a missed sell order to the table."""
        LOG.debug(
            "Adding unsold buy order txid to the 'unsold_buy_order_txids' table: %s",
            txid,
        )
        self.__db.add_row(
            self.__table,
            userref=self.__userref,
            txid=txid,
            price=price,
        )

    def remove(self: Self, txid: str) -> None:
        """Remove txid from the table."""
        LOG.debug(
            "Removing unsold buy order txid from the 'unsold_buy_order_txids'"
            " table: %s",
            txid,
        )
        self.__db.delete_row(
            self.__table,
            filters={
                "userref": self.__userref,
                "txid": txid,
            },
        )

    def get(self: Self, filters: dict | None = None) -> MappingResult:
        """Retrieve unsold buy order txids from the table."""
        LOG.debug(
            "Retrieving unsold buy order txids from the"
            " 'unsold_buy_order_txids' table with filters: %s",
            filters,
        )
        if not filters:
            filters = {}
        return self.__db.get_rows(
            self.__table,
            filters=filters | {"userref": self.__userref},
        )

    def count(self: Self, filters: dict | None = None) -> int:
        """Count unsold buy order txids from the table."""
        LOG.debug(
            "Count unsold buy order txids from the table unsold_buy_order_txids"
            " table with filters: %s",
            filters,
        )
        if not filters:
            filters = {}
        filters |= {"userref": self.__userref}

        query = (
            select(func.count())  # pylint: disable=not-callable
            .select_from(self.__table)
            .where(
                *(self.__table.c[column] == value for column, value in filters.items()),
            )
        )
        return self.__db.session.execute(query).scalar()  # type: ignore[no-any-return]


class PendingIXIDs:
    """
    Table containing pending TXIDs. TXIDs are pending for the time from being
    placed to processed by Kraken. Usually an order gets placed, the TXID is
    returned and stored in this table. Then the algorithm fetches this 'pending'
    TXID to retrieve the full order information in order to add these to the
    local orderbook. After that, the TXID gets removed from this table.
    """

    def __init__(self: Self, userref: int, db: DBConnect) -> None:
        LOG.debug("Initializing the PendingIXIDs table...")
        self.__db = db
        self.__userref = userref
        self.__table = Table(
            "pending_txids",
            self.__db.metadata,
            Column("id", Integer, primary_key=True),
            Column("userref", Integer, nullable=False),
            Column("txid", String, nullable=False),
        )

    def get(self: Self, filters: dict | None = None) -> MappingResult:
        """Get pending orders from the table."""
        LOG.debug(
            "Getting pending orders from the 'pending_txids' table with filters: %s",
            filters,
        )
        if not filters:
            filters = {}

        return self.__db.get_rows(
            self.__table,
            filters=filters | {"userref": self.__userref},
        )

    def add(self: Self, txid: str) -> None:
        """Add a pending order to the table."""
        LOG.debug(
            "Adding a pending txid to the 'pending_txids' table: %s",
            txid,
        )
        self.__db.add_row(
            self.__table,
            userref=self.__userref,
            txid=txid,
        )

    def remove(self: Self, txid: str) -> None:
        """Remove a pending order from the table."""
        LOG.debug("Removing pending txid from the 'pending_txids' table: %s", txid)
        self.__db.delete_row(
            self.__table,
            filters={"userref": self.__userref, "txid": txid},
        )

    def count(self: Self, filters: dict | None = None) -> int:
        """Count pending orders in the table."""
        LOG.debug(
            "Counting pending txids of the 'pending_txids' table with filters: %s",
            filters,
        )
        if not filters:
            filters = {}
        filters |= {"userref": self.__userref}

        query = (
            select(func.count())  # pylint: disable=not-callable
            .select_from(self.__table)
            .where(
                *(self.__table.c[column] == value for column, value in filters.items()),
            )
        )
        return self.__db.session.execute(query).scalar()  # type: ignore[no-any-return]



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2023 Benjamin Thomas Schwertfeger
# All rights reserved.
# https://github.com/btschwertfeger
#

from datetime import datetime
from logging import getLogger
from typing import TYPE_CHECKING, Self

import requests

if TYPE_CHECKING:

    from kraken_infinity_grid.gridbot import KrakenInfinityGridBot

LOG = getLogger(__name__)


class Telegram:
    """Telegram class to send messages to a Telegram chat."""

    def __init__(
        self: Self,
        strategy: "KrakenInfinityGridBot",
        telegram_token: str,
        telegram_chat_id: str,
        exception_token: str,
        exception_chat_id: str,
    ) -> None:
        self.__s = strategy
        self.__telegram_token = telegram_token
        self.__telegram_chat_id = telegram_chat_id
        self.__exception_token = exception_token
        self.__exception_chat_id = exception_chat_id

    def send_to_telegram(
        self: Self,
        message: str,
        exception: bool | None = False,
        log: bool = True,
    ) -> None:
        """Send a message to a Telegram chat"""
        if exception:
            if log:
                LOG.error(message)
            if not (self.__exception_token and self.__exception_chat_id):
                return
            response = requests.post(
                url=f"https://api.telegram.org/bot{self.__exception_token}/sendMessage",
                params={
                    "chat_id": self.__exception_chat_id,
                    "text": f"```\n{message}\n```",
                    "parse_mode": "markdown",
                },
                timeout=10,
            )
        else:
            if log:
                LOG.info(message)
            if not (self.__telegram_token and self.__telegram_chat_id):
                return
            response = requests.post(
                url=f"https://api.telegram.org/bot{self.__telegram_token}/sendMessage",
                params={
                    "chat_id": self.__telegram_chat_id,
                    "text": message,
                    "parse_mode": "markdown",
                },
                timeout=10,
            )

        if response.status_code != 200:
            # Its not that important to send telegram messages... so we just log
            # this here. The user will know that something is wrong when not
            # receiving regular status updates.
            LOG.error(
                "Failed to send message to Telegram. Status code: %d, message: \n%s",
                response.status_code,
                message,
            )

    def send_telegram_update(self: Self) -> None:
        """Send a message to the Telegram chat with the current status."""
        balances = self.__s.get_balances()

        message = f"👑 {self.__s.symbol}\n"
        message += f"└ Price » {self.__s.ticker.last} {self.__s.quote_currency}\n\n"

        message += "⚜️ Account\n"
        message += f"├ Total {self.__s.base_currency} » {balances['base_balance']}\n"
        message += f"├ Total {self.__s.quote_currency} » {balances['quote_balance']}\n"
        message += (
            f"├ Available {self.__s.quote_currency} » {balances['quote_available']}\n"
        )
        message += f"├ Available {self.__s.base_currency} » {balances['base_available'] - float(self.__s.configuration.get()['vol_of_unfilled_remaining'])}\n"  # noqa: E501
        message += f"├ Unfilled surplus of {self.__s.base_currency} » {self.__s.configuration.get()['vol_of_unfilled_remaining']}\n"  # noqa: E501
        message += f"├ Wealth » {round(balances['base_balance'] * self.__s.ticker.last + balances['quote_balance'], self.__s.cost_decimals)} {self.__s.quote_currency}\n"  # noqa: E501
        message += f"└ Investment » {round(self.__s.investment, self.__s.cost_decimals)} / {self.__s.max_investment} {self.__s.quote_currency}\n\n"  # noqa: E501

        message += "💠 Orders\n"
        message += f"├ Amount per Grid » {self.__s.amount_per_grid} {self.__s.quote_currency}\n"
        message += f"└ Open orders » {self.__s.orderbook.count()}\n"

        message += "\n```\n"
        message += f" 🏷️ Price in {self.__s.quote_currency}\n"
        max_orders_to_list: int = 5

        next_sells = [
            order["price"]
            for order in self.__s.orderbook.get_orders(
                filters={"side": "sell"},
                order_by=("price", "ASC"),
                limit=max_orders_to_list,
            )
        ]
        next_sells.reverse()

        if (n_sells := len(next_sells)) == 0:
            message += f"└───┬> {self.__s.ticker.last}\n"
        else:
            for index, sell_price in enumerate(next_sells):
                change = (sell_price / self.__s.ticker.last - 1) * 100
                if index == 0:
                    message += f" │  ┌[ {sell_price} (+{change:.2f}%)\n"
                elif index <= n_sells - 1 and index != max_orders_to_list:
                    message += f" │  ├[ {sell_price} (+{change:.2f}%)\n"
            message += f" └──┼> {self.__s.ticker.last}\n"

        next_buys = [
            order["price"]
            for order in self.__s.orderbook.get_orders(
                filters={"side": "buy"},
                order_by=("price", "DESC"),
                limit=max_orders_to_list,
            )
        ]
        if (n_buys := len(next_buys)) != 0:
            for index, buy_price in enumerate(next_buys):
                change = (buy_price / self.__s.ticker.last - 1) * 100
                if index < n_buys - 1 and index != max_orders_to_list:
                    message += f"    ├[ {buy_price} ({change:.2f}%)\n"
                else:
                    message += f"    └[ {buy_price} ({change:.2f}%)"
        message += "\n```"

        self.send_to_telegram(message)
        self.__s.configuration.update({"last_telegram_update": datetime.now()})



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2023 Benjamin Thomas Schwertfeger
# All rights reserved.
# https://github.com/btschwertfeger
#

"""Module that implements the main strategy"""

import asyncio
import sys
import traceback
from contextlib import suppress
from datetime import datetime, timedelta
from decimal import Decimal
from importlib.metadata import version
from logging import getLogger
from time import sleep
from types import SimpleNamespace
from typing import Iterable, Optional, Self

from kraken.exceptions import (
    KrakenAuthenticationError,
    KrakenInvalidOrderError,
    KrakenPermissionDeniedError,
)
from kraken.spot import Market, SpotWSClient, Trade, User

from kraken_infinity_grid.database import (
    Configuration,
    DBConnect,
    Orderbook,
    PendingIXIDs,
    UnsoldBuyOrderTXIDs,
)
from kraken_infinity_grid.order_management import OrderManager
from kraken_infinity_grid.setup import SetupManager
from kraken_infinity_grid.telegram import Telegram

LOG = getLogger(__name__)


class KrakenInfinityGridBot(SpotWSClient):
    """
    The KrakenInfinityGridBot class implements the infinity grid trading bot
    strategy.

    The bot is designed to trade on the Kraken exchange and uses the Kraken API
    to place orders and fetch information about the current trading pair.

    All actions are triggered via websocket messages received by the
    ``on_message`` function, which is a callback function used by the
    python-kraken-sdk.

    The algorithm manages its orders in an orderbook, configuration, and pending
    transactions in separate tables in a SQL-based database (PostgreSQL). It
    uses the passed ``userref`` to distinguish between different instances.
    E.g.: For one instance of the algorithm one uses the userref 1680953420 for
    the DOT/USD pair, and 1680953421 for the BTC/USD pair.

    The available strategies are documented in the README.md file of this
    project.

    Internal
    ========

    The following steps are a rough description of the internal workflow of the
    algorithm.

    Initialization:

    - This class is derived from the python-kraken-sdk's websocket client, which
      allows for subscribing to channels and receiving messages. During the
      init, the algorithm will establish the connection to the database as well
      as initializing the user, trade, and market clients.
    - The after the bot is created, events that came from subscribed websocket
      feeds will trigger the ``on_message`` function. Subscribed feeds are the
      ticker (of the asset pair) and execution (for order execution updates).
    - If both channels are connected, the algorithm does an initial setup:
        - Check pre-conditions
        - Validate the configuration
        - Retrieve information about the trading pair (fee, min order size,
          etc.)
        - Check, update, and sync the local orderbook with upstream while
          recovering orders that might not yet found their way into the
          orderbook after placing.
        - If there are executed buy orders during downtime, the corresponding
          sell orders will be placed (depending on the strategy in use).
        - If all of these are done, the bot is ready to handle new incoming
          messages.

    New ticker message:

    - If the init is not done yet, the algorithm ignores the message.
    - The ticker will be updated.
    - The missing sell orders will be assigned (if any). This is done at this
      place in order to have a frequent check for missing sell orders.
    - The ``check_price_range`` function will be triggered.
        - This either calls ``assign_all_pending_txids`` to add orders that were
          placed but are not yet added to the local orderbook.
        - Or checks the current price range. The price range check is skipped in
          case of pending transactions to avoid double orders.
            - Ensure the existence $n$ open buy orders.
            - Cancel the lowest buy order if more than $n$ buy orders are open.
            - Shift-up buy orders if price rises to high.
                - Cancel all open buy orders
                - Calls ``check_price_range``
            - If strategy==SWING: Place sell orders if possible

    New execution message - new order:

    - If the init is not done yet, the algorithm ignores the message.
    - The algorithm ties to assign the new order to the orderbook. In some cases
      this fails, as newly placed orders may not be fetchable via REST API for
      some time. For this reason, there are some retries implemented.

    New execution message - filled order:

    - If the init is not done yet, the algorithm ignores the message.
    - If the filled order was a buy order (depending on the strategy), the
      algorithm places a sell order and updates the local orderbook.
    - If the filled order was a sell order (depending on the strategy), the
      algorithm places a buy order and updates the local orderbook.

    New execution message - cancelled order:

    - If the init is not done yet, the algorithm ignores the message.
    - The algorithm removes the order from the local orderbook and ensures that
      in case of a partly filled order, the remaining volume is saved and placed
      as sell order somewhen later (if it was a buy order). Sell orders usually
      don't get cancelled.

    """

    def __init__(  # pylint: disable=too-many-arguments
        self: Self,
        key: str,
        secret: str,
        config: dict,
        db_config: dict,
        dry_run: bool = False,
    ) -> None:
        super().__init__(key=key, secret=secret)

        LOG.info(
            "Initiate the Kraken Infinity Grid Algorithm instance (v%s)",
            version("kraken-infinity-grid"),
        )
        LOG.debug("Config: %s", config)
        self.init_done: bool = False
        self.dry_run: bool = dry_run

        # Settings and config collection
        ##
        self.strategy: str = config["strategy"]
        self.userref: int = config["userref"]
        self.name: str = config["name"]

        # Commonly used config values
        ##
        self.interval: float = float(config["interval"])
        self.amount_per_grid: float = float(config["amount_per_grid"])

        self.amount_per_grid_plus_fee: float | None = config.get("fee")

        self.max_investment: float = config["max_investment"]
        self.n_open_buy_orders: int = config["n_open_buy_orders"]
        self.fee: float | None = config.get("fee")
        self.base_currency: str = config["base_currency"]
        self.quote_currency: str = config["quote_currency"]

        self.symbol: str = self.base_currency + "/" + self.quote_currency  # BTC/EUR
        self.xsymbol: str | None = None  # XXBTZEUR
        self.altname: str | None = None  # XBTEUR
        self.zbase_currency: str | None = None  # XXBT
        self.xquote_currency: str | None = None  # ZEUR
        self.cost_decimals: int | None = None  # 5 for EUR, i.e., 0.00001 EUR

        # If the algorithm receives execution messages before being ready to
        # trade, they will be stored here and processed later.
        ##
        self.__missed_messages: list[dict] = []

        # Wait for ticker to be connected before do some action
        ##
        self.__ticker_channel_connected: bool = False
        self.ticker: SimpleNamespace = None

        self.is_ready_to_trade: bool = False
        self.__execution_channel_connected: bool = False

        # Define the Kraken clients
        ##
        self.user: User = User(key=key, secret=secret)
        self.market: Market = Market(key=key, secret=secret)
        self.trade: Trade = Trade(key=key, secret=secret)

        # Instantiate the algorithm's components
        ##
        self.om = OrderManager(strategy=self)
        self.sm = SetupManager(strategy=self)
        self.t = Telegram(
            strategy=self,
            telegram_token=config["telegram_token"],
            telegram_chat_id=config["telegram_chat_id"],
            exception_token=config["exception_token"],
            exception_chat_id=config["exception_chat_id"],
        )

        # Database setup
        ##
        self.database: DBConnect = DBConnect(**db_config)
        self.orderbook: Orderbook = Orderbook(userref=self.userref, db=self.database)
        self.configuration: Configuration = Configuration(
            userref=self.userref,
            db=self.database,
        )
        self.pending_txids: PendingIXIDs = PendingIXIDs(
            userref=self.userref,
            db=self.database,
        )
        self.unsold_buy_order_txids: UnsoldBuyOrderTXIDs = UnsoldBuyOrderTXIDs(
            userref=self.userref,
            db=self.database,
        )
        self.database.init_db()

    async def on_message(  # noqa: C901, PLR0912
        self: Self,
        message: dict | list,
    ) -> None:
        """
        This function receives all messages that are sent via the websocket
        connections by Kraken. It's the entrypoint of the incoming messages and
        calls the appropriate functions to handle the messages.
        """
        try:
            # =====================================================================
            # Filtering out unwanted messages
            if not isinstance(message, dict):
                LOG.warning("Message is not a dict: %s", message)
                return

            if (channel := message.get("channel")) in {"heartbeat", "status"}:
                return

            if message.get("method"):
                if message["method"] == "subscribe" and not message["success"]:
                    self.save_exit(
                        "The algorithm was not able to subscribe to selected"
                        " channels. Please check the logs.",
                    )
                return

            # =====================================================================
            # Initial setup
            if channel == "ticker" and not self.__ticker_channel_connected:
                self.__ticker_channel_connected = True
                # Set ticker the first time to have the ticker set during setup.
                self.ticker = SimpleNamespace(last=float(message["data"][0]["last"]))
                LOG.info("- Subscribed to ticker channel successfully!")

            elif channel == "executions" and not self.__execution_channel_connected:
                self.__execution_channel_connected = True
                LOG.info("- Subscribed to execution channel successfully!")

            if (
                self.__ticker_channel_connected
                and self.__execution_channel_connected
                and not self.is_ready_to_trade
            ):
                # Sets self.is_ready_to_trade to True
                self.sm.prepare_for_trading()

                # If there are any missed messages, process them now.
                for msg in self.__missed_messages:
                    await self.on_message(msg)
                self.__missed_messages = []

            if not self.is_ready_to_trade:
                if channel == "executions":
                    # If the algorithm is not ready to trade, store the
                    # executions to process them later.
                    self.__missed_messages.append(message)

                # Return here, until the algorithm is ready to trade. It is
                # ready when the init/setup is done and the orderbook is
                # updated.
                return

            # =====================================================================
            # Handle ticker and execution messages

            if (
                channel == "ticker"
                and (data := message.get("data"))
                and data[0].get("symbol") == self.symbol
            ):
                self.configuration.update({"last_price_time": datetime.now()})

                self.ticker = SimpleNamespace(last=float(data[0]["last"]))
                if self.unsold_buy_order_txids.count() != 0:
                    self.om.add_missed_sell_orders()

                self.om.check_price_range()

            elif channel == "executions" and (data := message.get("data", [])):
                if message.get("type") == "snapshot":
                    # Snapshot data is not interesting, as this is handled
                    # during sync with upstream.
                    return

                for execution in data:
                    LOG.debug("Got execution: %s", execution)
                    match execution["exec_type"]:
                        case "new":
                            self.om.assign_order_by_txid(execution["order_id"])
                        case "filled":
                            self.om.handle_filled_order_event(execution["order_id"])
                        case "canceled" | "expired":
                            self.om.handle_cancel_order(execution["order_id"])

        except Exception as exc:  # noqa: BLE001
            LOG.error(msg="Exception while processing message.", exc_info=exc)
            self.save_exit(reason=traceback.format_exc())

    # ==========================================================================

    async def run(self: Self) -> None:
        """
        Main function that starts the algorithm and runs it until it is
        interrupted.
        """
        LOG.info("Starting the Kraken Infinity Grid Algorithm...")

        # ======================================================================
        # Try to connect to the Kraken API, validate credentials and API key
        # permissions.
        ##
        self.__check_kraken_status()

        try:
            self.__check_api_keys()
        except (KrakenAuthenticationError, KrakenPermissionDeniedError) as exc:
            await self.close()  # Stops the websocket connections and aiohttp session
            self.save_exit(
                (
                    "Passed API keys are invalid!"
                    if isinstance(exc, KrakenAuthenticationError)
                    else "Passed API keys are missing permissions!"
                ),
            )

        # ======================================================================
        # Start the websocket connections and run the main function
        ##
        try:
            await self.__main()
        except asyncio.CancelledError:
            await self.close()  # Stops the websocket connections and aiohttp session
            self.save_exit("The algorithm was interrupted!")

    async def __main(self: Self) -> None:
        """
        Main function that runs the algorithm. It subscribes to the ticker and
        execution channels and runs the main loop until the algorithm is
        interrupted.
        """

        # ======================================================================
        # Start the websocket connection
        ##
        LOG.info("Starting the websocket connection...")
        await self.start()
        LOG.info("Websocket connection established!")

        # ======================================================================
        # Subscribe to the execution and ticker channels
        ##
        LOG.info("Subscribing to channels...")
        await self.subscribe(
            params={
                "channel": "ticker",
                "symbol": [f"{self.base_currency}/{self.quote_currency}"],
            },
        )
        await self.subscribe(
            {
                "channel": "executions",
                # Snapshots are only required to check if the channel is
                # connected. They are not used for any other purpose.
                "snap_orders": True,
                "snap_trades": True,
            },
        )

        # Set this initially in case the DB contains a value that is too old.
        self.configuration.update({"last_price_time": datetime.now()})

        # ======================================================================
        # Main Loop: Run until interruption
        ##
        # 'self.exception_occur' is how the python-kraken-sdk notifies about
        # exceptions in the websocket connection.
        while not self.exception_occur:
            try:
                conf = self.configuration.get()
                now = datetime.now()
                last_hour = now - timedelta(hours=1)

                if self.init_done and (
                    not conf["last_price_time"]
                    or not conf["last_telegram_update"]
                    or conf["last_telegram_update"] < last_hour
                ):
                    # Send update once per hour to Telegram
                    self.t.send_telegram_update()

                if conf["last_price_time"] + timedelta(seconds=600) < now:
                    # Exit if no price update for a long time (10 minutes).
                    self.save_exit(
                        reason="No price update for a long time, exit!",
                    )

            except (
                Exception  # pylint: disable=broad-exception-caught # noqa: BLE001
            ) as exc:
                self.save_exit(
                    reason=f"Exception in main: {exc} {traceback.format_exc()}",
                )

            await asyncio.sleep(6)

        self.save_exit(
            reason="The websocket connection encountered an exception!",
        )

    def save_exit(self: Self, reason: str = "") -> None:
        """Save exit triggered, saving data and exit the program."""
        message: str = str(
            f"❌ {self.name} ❌"
            f"\n{self.symbol} Save exit triggered, saving data."
            f"\nReason: {reason}",
        )

        try:
            self.t.send_to_telegram(message=message, exception=True)
        except Exception as exc:  # noqa: BLE001
            LOG.warning("%s", message)
            LOG.error("FIXME: look at this: %s: %s", exc, traceback.format_exc())

        sys.exit(1)

    def __check_kraken_status(self: Self, tries: int = 0) -> None:
        """Checks whether the Kraken API is available."""
        if tries == 3:
            LOG.error("- Could not connect to the Kraken Exchange API.")
            sys.exit(1)
        try:
            self.market.get_system_status()
            LOG.info("- Kraken Exchange API Status: Available")
        except (
            Exception  # pylint: disable=broad-exception-caught # noqa: BLE001
        ) as exc:
            LOG.debug(
                "Exception while checking Kraken availability {exc} {traceback}",
                extra={"exc": exc, "traceback": traceback.format_exc()},
            )
            LOG.warning("- Kraken not available. (Try %d/3)", tries + 1)
            sleep(3)
            self.__check_kraken_status(tries=tries + 1)

    def __check_api_keys(self: Self) -> None:
        """
        Checks if the credentials are valid and if the API keys have the
        required permissions.
        """
        LOG.info("- Checking permissions of API keys...")

        LOG.info(" - Checking if 'Query Funds' permission set...")
        self.user.get_account_balance()

        LOG.info(" - Checking if 'Query open order & trades' permission set...")
        self.user.get_open_orders(trades=True)

        LOG.info(" - Checking if 'Query closed order & trades' permission set...")
        self.user.get_closed_orders(trades=True)

        LOG.info(" - Checking if 'Create & modify orders' permission set...")
        self.trade.create_order(
            pair=self.symbol,
            side="buy",
            ordertype="market",
            volume="0.0001",
            price="1",
            validate=True,
        )
        LOG.info(" - Checking if 'Cancel & close orders' permission set...")
        with suppress(KrakenInvalidOrderError):
            self.trade.cancel_order(
                txid="",
                extra_params={"cl_ord_id": "kraken_infinity_grid_internal"},
            )

        LOG.info(" - Checking if 'Websocket interface' permission set...")
        self.trade.request(
            method="POST",
            uri="/0/private/GetWebSocketsToken",
        )

        LOG.info(" - Passed API keys and permissions are valid!")

    # ======================================================================
    # Helper Functions
    ##
    def get_balances(self: Self) -> dict[str, float]:
        """
        Returns the available and overall balances of the quote and base
        currency.
        """
        LOG.debug("Retrieving the user's balances...")

        base_balance: Decimal = Decimal(0)
        base_available: Decimal = Decimal(0)
        quote_balance: Decimal = Decimal(0)
        quote_available: Decimal = Decimal(0)

        for symbol, data in self.user.get_balances().items():
            if symbol == self.zbase_currency:
                base_balance = Decimal(data["balance"])
                base_available = base_balance - Decimal(data["hold_trade"])
            elif symbol == self.xquote_currency:
                quote_balance = Decimal(data["balance"])
                quote_available = quote_balance - Decimal(data["hold_trade"])

        balances = {
            "base_balance": float(base_balance),
            "quote_balance": float(quote_balance),
            "base_available": float(base_available),
            "quote_available": float(quote_available),
        }
        LOG.debug("Retrieved balances: %s", balances)
        return balances

    def get_current_buy_prices(self: Self) -> Iterable[float]:
        """Returns a generator of the prices of open buy orders."""
        LOG.debug("Getting current buy prices...")
        for order in self.orderbook.get_orders(filters={"side": "buy"}):
            yield order["price"]

    def get_order_price(
        self: Self,
        side: str,
        last_price: float,
        extra_sell: Optional[bool] = False,
    ) -> float:
        """
        Returns the order price depending on the strategy and side. Also assigns
        a new highest buy price to configuration if there was a new highest
        buy.
        """
        LOG.debug("Computing the order price...")
        order_price: float
        price_of_highest_buy = self.configuration.get()["price_of_highest_buy"]
        last_price = float(last_price)

        if side == "sell":  # New order is a sell
            if self.strategy == "SWING" and extra_sell:
                # Extra sell order when SWING
                # 2x interval above [last close price | price of highest buy]
                order_price = last_price * (1 + self.interval) * (1 + self.interval)
                if order_price < price_of_highest_buy:
                    order_price = (
                        price_of_highest_buy * (1 + self.interval) * (1 + self.interval)
                    )

            else:
                # Regular sell order (even for SWING) (cDCA will trigger this
                # but it will be filtered out later)
                if last_price > price_of_highest_buy:
                    self.configuration.update(
                        {"price_of_highest_buy": last_price},
                    )

                # Sell price 1x interval above buy price
                order_price = last_price * (1 + self.interval)
                if self.ticker.last > order_price:
                    order_price = self.ticker.last * (1 + self.interval)
            return order_price

        if side == "buy":  # New order is a buy
            order_price = last_price * 100 / (100 + 100 * self.interval)
            if order_price > self.ticker.last:
                order_price = self.ticker.last * 100 / (100 + 100 * self.interval)
            return order_price

        raise ValueError(f"Unknown side: {side}!")

    def get_value_of_orders(self: Self, orders: Iterable) -> float:
        """Returns the overall invested quote that is invested"""
        LOG.debug("Getting value of open orders...")
        investment = sum(
            float(order["price"]) * float(order["volume"]) for order in orders
        )
        LOG.debug("Value of open orders: %d %s", investment, self.quote_currency)
        return investment

    @property
    def investment(self: Self) -> float:
        """Returns the current investment based on open orders."""
        return self.get_value_of_orders(
            orders=self.orderbook.get_orders(),
        )

    @property
    def max_investment_reached(self: Self) -> bool:
        """Returns True if the maximum investment is reached."""
        return (
            self.max_investment <= self.investment + self.amount_per_grid_plus_fee
        ) or (self.max_investment <= self.investment)



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2023 Benjamin Thomas Schwertfeger
# All rights reserved.
# https://github.com/btschwertfeger
#

""" Order management for the Kraken Infinity Grid Bot. """

from __future__ import annotations

import logging
from decimal import Decimal
from time import sleep
from typing import TYPE_CHECKING, Self

from kraken.exceptions import KrakenUnknownOrderError

if TYPE_CHECKING:
    # To avoid circular import for type checking
    from kraken_infinity_grid.gridbot import KrakenInfinityGridBot

LOG: logging.Logger = logging.getLogger(__name__)


class OrderManager:
    """Manages the orderbook and the order handling."""

    def __init__(self: OrderManager, strategy: KrakenInfinityGridBot) -> None:
        LOG.debug("Initializing the OrderManager...")
        self.__s = strategy

    def add_missed_sell_orders(self: Self) -> None:
        """
        This functions can create sell orders in case there is at least one
        executed buy order that is missing its sell order.

        Missed sell orders came into place when a buy was executed and placing
        the sell failed. An entry to the missed sell order id table is added
        right before placing a sell order.
        """
        LOG.info("- Create sell orders based on unsold buy orders...")
        for entry in self.__s.unsold_buy_order_txids.get():
            LOG.info("  - %s", entry)
            self.handle_arbitrage(
                side="sell",
                order_price=entry["price"],
                txid_to_delete=entry["txid"],
            )

    def assign_all_pending_transactions(self: Self) -> None:
        """Assign all pending transactions to the orderbook."""
        LOG.info("- Checking pending transactions...")
        for order in self.__s.pending_txids.get():
            self.assign_order_by_txid(txid=order["txid"])

    def assign_order_by_txid(self: Self, txid: str) -> None:
        """
        Assigns an order by its txid to the orderbook.

        - Option 1: Removes them from the pending txids and appends it to
                    the orderbook
        - Option 2: Updates the info of the order in the orderbook

        There is no need for checking the order status, since after the order
        was added to the orderbook, the algorithm will handle any removals in
        case of closed orders.
        """
        LOG.info("Processing '%s' ...", txid)
        order_details = self.get_orders_info_with_retry(txid=txid)
        LOG.debug("- Order information: %s", order_details)

        if (
            order_details["descr"]["pair"] != self.__s.altname
            or order_details["userref"] != self.__s.userref
        ):
            LOG.info("Order '%s' does not belong to this instance.", txid)
            return

        if self.__s.pending_txids.count(filters={"txid": order_details["txid"]}) != 0:
            self.__s.orderbook.add(order_details)
            self.__s.pending_txids.remove(order_details["txid"])
        else:
            self.__s.orderbook.update(
                order_details,
                filters={"txid": order_details["txid"]},
            )
            LOG.info("Updated order '%s' in orderbook.", order_details["txid"])

        LOG.info(
            "Current invested value: %f / %d %s",
            self.__s.investment,
            self.__s.max_investment,
            self.__s.quote_currency,
        )

    # ==========================================================================
    #            C H E C K - P R I C E - R A N G E
    # ==========================================================================

    def __check_pending_txids(self: OrderManager) -> bool:
        """
        Skip checking the price range, because first all missing orders
        must be assigned. Otherwise this could lead to double trades.

        Returns False if okay and True if ``check_price_range`` must be skipped.
        """
        if self.__s.pending_txids.count() != 0:
            LOG.info("check_price_range... skip because pending_txids != 0")
            self.assign_all_pending_transactions()
            return True
        return False

    def __check_near_buy_orders(self: OrderManager) -> None:
        """
        Cancel buy orders that are next to each other. Only the lowest buy order
        will survive. This is to avoid that the bot buys at the same price
        multiple times.

        Other functions handle the eventual cancellation of a very low buy order
        to avoid falling out of the price range.
        """
        LOG.debug("Checking if distance between buy orders is too low...")

        if len(buy_prices := list(self.__s.get_current_buy_prices())) == 0:
            return

        buy_prices.sort(reverse=True)
        for i, price in enumerate(buy_prices[1:]):
            if (
                price == buy_prices[i]
                or (buy_prices[i] / price) - 1 < self.__s.interval / 2
            ):
                for order in self.__s.orderbook.get_orders(filters={"side": "buy"}):
                    if order["price"] == buy_prices[i]:
                        self.handle_cancel_order(txid=order["txid"])
                        break

    def __check_n_open_buy_orders(self: OrderManager) -> None:
        """
        Ensures that there are n open buy orders and will place orders until n.
        """
        LOG.debug(
            "Checking if there are %d open buy orders...",
            self.__s.n_open_buy_orders,
        )
        can_place_buy_order: bool = True
        buy_prices: list[float] = list(self.__s.get_current_buy_prices())

        while (
            (n_active_buy_orders := self.__s.orderbook.count(filters={"side": "buy"}))
            < self.__s.n_open_buy_orders
            and can_place_buy_order
            and self.__s.pending_txids.count() == 0
            and not self.__s.max_investment_reached
        ):
            fetched_balances: dict[str, float] = self.__s.get_balances()
            if fetched_balances["quote_available"] > self.__s.amount_per_grid_plus_fee:
                order_price: float = self.__s.get_order_price(
                    side="buy",
                    last_price=(
                        self.__s.ticker.last
                        if n_active_buy_orders == 0
                        else min(buy_prices)
                    ),
                )

                self.handle_arbitrage(side="buy", order_price=order_price)
                buy_prices = list(self.__s.get_current_buy_prices())
                LOG.info("Length of active buy orders: %s", n_active_buy_orders + 1)
            else:
                LOG.warning("Not enough quote currency available to place buy order!")
                can_place_buy_order = False

    def __check_lowest_cancel_of_more_than_n_buy_orders(self: OrderManager) -> None:
        """
        Cancel the lowest buy order if new higher buy was placed because of an
        executed sell order.
        """
        LOG.debug("Checking if the lowest buy order needs to be canceled...")

        if (
            n_to_cancel := (
                self.__s.orderbook.count(filters={"side": "buy"})
                - self.__s.n_open_buy_orders
            )
        ) > 0:
            for order in self.__s.orderbook.get_orders(
                filters={"side": "buy"},
                order_by=("price", "asc"),
                limit=n_to_cancel,
            ):
                self.handle_cancel_order(txid=order["txid"])

    def __shift_buy_orders_up(self: OrderManager) -> bool:
        """
        Checks if the buy order prices are not to low. If there are too low,
        they get canceled and the ``check_price_range`` function is triggered
        again to place new buy orders.

        Returns ``True`` if the orders get canceled and the
        ``check_price_range`` functions stops.
        """
        LOG.debug("Checking if buy orders need to be shifted up...")

        if (
            max_buy_order := self.__s.orderbook.get_orders(
                filters={"side": "buy"},
                order_by=("price", "desc"),
                limit=1,
            ).first()  # type: ignore[no-untyped-call]
        ) and (
            self.__s.ticker.last
            > max_buy_order["price"]
            * (1 + self.__s.interval)
            * (1 + self.__s.interval)
            * 1.001
        ):
            self.cancel_all_open_buy_orders()
            self.check_price_range()
            return True

        return False

    def __check_extra_sell_order(self: OrderManager) -> None:
        """
        Checks if an extra sell order can be placed. This only applies for the
        SWING strategy.
        """
        if self.__s.strategy != "SWING":
            return

        LOG.debug("Checking if extra sell order can be placed...")
        if self.__s.orderbook.count(filters={"side": "sell"}) == 0:
            fetched_balances = self.__s.get_balances()

            if (
                fetched_balances["base_available"] * self.__s.ticker.last
                > self.__s.amount_per_grid_plus_fee
            ):
                order_price = self.__s.get_order_price(
                    side="sell",
                    last_price=self.__s.ticker.last,
                    extra_sell=True,
                )
                self.__s.t.send_to_telegram(
                    f"ℹ️ {self.__s.symbol}: Placing extra sell order",  # noqa: RUF001
                )
                self.handle_arbitrage(side="sell", order_price=order_price)

    def check_price_range(self: OrderManager) -> None:
        """
        Checks if the orders prices match the conditions of the bot respecting
        the current price.

        If the price (``self.ticker.last``) raises to high, the open buy orders
        will be canceled and new buy orders below the price respecting the
        interval will be placed.
        """
        if self.__s.dry_run:
            LOG.debug("Dry run, not checking price range.")
            return

        LOG.debug("Check conditions for upgrading the grid...")

        if self.__check_pending_txids():
            LOG.debug("Not checking price range because of pending txids.")
            return

        # Remove orders that are next to each other
        self.__check_near_buy_orders()

        # Ensure n open buy orders
        self.__check_n_open_buy_orders()

        # Return if some newly placed order is still pending and not in the
        # orderbook.
        if self.__s.pending_txids.count() != 0:
            return

        # Check if there are more than n buy orders and cancel the lowest
        self.__check_lowest_cancel_of_more_than_n_buy_orders()

        # Check the price range and shift the orders up if required
        if self.__shift_buy_orders_up():
            return

        # Place extra sell order (only for SWING strategy)
        self.__check_extra_sell_order()

    # =============================================================================
    #           C R E A T E / C A N C E L - O R D E R S
    # =============================================================================

    def handle_arbitrage(
        self: Self,
        side: str,
        order_price: float,
        txid_to_delete: str | None = None,
    ) -> None:
        """
        Handles the arbitrage between buy and sell orders.

        The existence of this function is mainly justified due to the sleep
        statement at the end.
        """
        LOG.debug(
            "Handle arbitrage for %s order with order price: %s and"
            " txid_to_delete: %s",
            side,
            order_price,
            txid_to_delete,
        )

        if self.__s.dry_run:
            LOG.info("Dry run, not placing %s order.", side)
            return

        if side == "buy":
            self.new_buy_order(
                order_price=order_price,
                txid_to_delete=txid_to_delete,
            )
        elif side == "sell":
            self.new_sell_order(
                order_price=order_price,
                txid_to_delete=txid_to_delete,
            )

        # Wait a bit to avoid rate limiting.
        sleep(0.2)

    def new_buy_order(
        self: OrderManager,
        order_price: float,
        txid_to_delete: str | None = None,
    ) -> None:
        """Places a new buy order."""
        if self.__s.dry_run:
            LOG.info("Dry run, not placing buy order.")
            return

        if txid_to_delete is not None:
            self.__s.orderbook.remove(filters={"txid": txid_to_delete})

        if (
            self.__s.orderbook.count(filters={"side": "buy"})
            >= self.__s.n_open_buy_orders
        ):
            # Don't place new buy orders if there are already enough
            return

        # Check if algorithm reached the max_investment value
        if self.__s.max_investment_reached:
            return

        # Compute the target price for the upcoming buy order.
        order_price = float(
            self.__s.trade.truncate(
                amount=order_price,
                amount_type="price",
                pair=self.__s.symbol,
            ),
        )

        # Compute the target volume for the upcoming buy order.
        # NOTE: The fee is respected while placing the sell order
        volume = float(
            self.__s.trade.truncate(
                amount=Decimal(self.__s.amount_per_grid) / Decimal(order_price),
                amount_type="volume",
                pair=self.__s.symbol,
            ),
        )

        # ======================================================================
        # Check if there is enough quote balance available to place a buy order.
        current_balances = self.__s.get_balances()
        if current_balances["quote_available"] > self.__s.amount_per_grid_plus_fee:
            LOG.info(
                "Placing order to buy %s %s @ %s %s.",
                volume,
                self.__s.base_currency,
                order_price,
                self.__s.quote_currency,
            )

            # Place a new buy order, append txid to pending list and delete
            # corresponding sell order from local orderbook.
            placed_order = self.__s.trade.create_order(
                ordertype="limit",
                side="buy",
                volume=volume,
                pair=self.__s.symbol,
                price=order_price,
                userref=self.__s.userref,
                validate=self.__s.dry_run,
                oflags="post",  # post-only buy orders
            )

            self.__s.pending_txids.add(placed_order["txid"][0])
            self.__s.om.assign_order_by_txid(placed_order["txid"][0])
            return

        # ======================================================================
        # Not enough available funds to place a buy order.
        message = f"⚠️ {self.__s.symbol}"
        message += f"├ Not enough {self.__s.quote_currency}"
        message += f"├ to buy {volume} {self.__s.base_currency}"
        message += f"└ for {order_price} {self.__s.quote_currency}"
        self.__s.t.send_to_telegram(message)
        LOG.warning("Current balances: %s", current_balances)
        return

    def new_sell_order(  # noqa: C901
        self: OrderManager,
        order_price: float,
        txid_to_delete: str | None = None,
    ) -> None:
        """Places a new sell order."""
        if self.__s.dry_run:
            LOG.info("Dry run, not placing sell order.")
            return

        if self.__s.strategy == "cDCA":
            LOG.debug("cDCA strategy, not placing sell order.")
            if txid_to_delete is not None:
                self.__s.orderbook.remove(filters={"txid": txid_to_delete})
            return

        LOG.debug("Check conditions for placing a sell order...")

        # ======================================================================
        volume: float | None = None
        if txid_to_delete is not None:  # If corresponding buy order filled
            # GridSell always has txid_to_delete set.

            # Add the txid of the corresponding buy order to the unsold buy
            # order txids in order to ensure that the corresponding sell order
            # will be placed - even if placing now fails.
            if not self.__s.unsold_buy_order_txids.get(
                filters={"txid": txid_to_delete},
            ).first():  # type: ignore[no-untyped-call]
                self.__s.unsold_buy_order_txids.add(
                    txid=txid_to_delete,
                    price=order_price,
                )

            # ==================================================================
            # Get the corresponding buy order in order to retrieve the volume.
            corresponding_buy_order = self.get_orders_info_with_retry(
                txid=txid_to_delete,
            )

            # In some cases the corresponding buy order is not closed yet and
            # the vol_exec is missing. In this case, the function will be
            # called again after a short delay.
            if (
                corresponding_buy_order["status"] != "closed"
                or corresponding_buy_order["vol_exec"] == 0
            ):
                LOG.warning(
                    "Can't place sell order, since the corresponding buy order"
                    " is not closed yet. Retry in 1 second. (order: %s)",
                    corresponding_buy_order,
                )
                sleep(1)
                self.__s.om.new_sell_order(
                    order_price=order_price,
                    txid_to_delete=txid_to_delete,
                )
                return

            if self.__s.strategy == "GridSell":
                # Volume of a GridSell is fixed to the executed volume of the
                # buy order.
                volume = float(
                    self.__s.trade.truncate(
                        amount=float(corresponding_buy_order["vol_exec"]),
                        amount_type="volume",
                        pair=self.__s.symbol,
                    ),
                )

        order_price = float(
            self.__s.trade.truncate(
                amount=order_price,
                amount_type="price",
                pair=self.__s.symbol,
            ),
        )

        if self.__s.strategy in {"GridHODL", "SWING"} or (
            self.__s.strategy == "GridSell" and volume is None
        ):
            # For GridSell: This is only the case if there is no corresponding
            # buy order and the sell order was placed, e.g. due to an extra sell
            # order via selling of partially filled buy orders.

            # Respect the fee to not reduce the quote currency over time, while
            # accumulating the base currency.
            volume = float(
                self.__s.trade.truncate(
                    amount=Decimal(self.__s.amount_per_grid)
                    / (Decimal(order_price) * (1 - (2 * Decimal(self.__s.fee)))),
                    amount_type="volume",
                    pair=self.__s.symbol,
                ),
            )

        # ======================================================================
        # Check if there is enough base currency available for selling.
        fetched_balances = self.__s.get_balances()
        if fetched_balances["base_available"] >= volume:
            # Place new sell order, append id to pending list, and delete
            # corresponding buy order from local orderbook.
            LOG.info(
                "Placing order to sell %s %s @ %s %s.",
                volume,
                self.__s.base_currency,
                order_price,
                self.__s.quote_currency,
            )

            placed_order = self.__s.trade.create_order(
                ordertype="limit",
                side="sell",
                volume=volume,
                pair=self.__s.symbol,
                price=order_price,
                userref=self.__s.userref,
                validate=self.__s.dry_run,
            )

            placed_order_txid = placed_order["txid"][0]
            self.__s.pending_txids.add(placed_order_txid)

            if txid_to_delete is not None:
                # Other than with buy orders, we can only delete the
                # corresponding buy order if the sell order was placed.
                self.__s.orderbook.remove(filters={"txid": txid_to_delete})
                self.__s.unsold_buy_order_txids.remove(txid=txid_to_delete)

            self.__s.om.assign_order_by_txid(txid=placed_order_txid)
            return

        # ======================================================================
        # Not enough funds to sell
        message = f"⚠️ {self.__s.symbol}"
        message += f"├ Not enough {self.__s.base_currency}"
        message += f"├ to sell {volume} {self.__s.base_currency}"
        message += f"└ for {order_price} {self.__s.quote_currency}"

        self.__s.t.send_to_telegram(message)
        LOG.warning("Current balances: %s", fetched_balances)

        if self.__s.strategy == "GridSell":
            # Restart the algorithm if there is not enough base currency to
            # sell. This could only happen if some orders have not being
            # processed properly, the algorithm is not in sync with the
            # exchange, or manual trades have been made during processing.
            self.__s.save_exit(reason=message)
        elif txid_to_delete is not None:
            # TODO: Check if this is appropriate or not
            #       Added logging statement to monitor occurrences
            # ... This would only be the case for GridHODL and SWING, while
            # those should always have enough base currency available... but
            # lets check this for a while.
            LOG.warning(
                "TODO: Not enough funds to place sell order for txid %s",
                txid_to_delete,
            )
            self.__s.orderbook.remove(filters={"txid": txid_to_delete})

    def handle_filled_order_event(
        self: OrderManager,
        txid: str,
    ) -> None:
        """
        Gets triggered by a filled order event from the ``on_message`` function.

        It fetches the filled order info (using some tries).

        If there is the KeyError which happens due to Krakens shitty, then wait
        for one second and this function will call it self again and return.
        """
        LOG.debug("Handling a new filled order event for txid: %s", txid)

        # ======================================================================
        # Fetch the order details for the given txid.
        ##
        order_details = self.get_orders_info_with_retry(txid=txid)

        # ======================================================================
        # Check if the order belongs to this bot and return if not
        ##
        if (
            order_details["descr"]["pair"] != self.__s.altname
            or order_details["userref"] != self.__s.userref
        ):
            LOG.debug(
                "Filled order %s was not from this bot or pair.",
                txid,
            )
            return

        # ======================================================================
        # Sometimes the order is not closed yet, so retry fetching the order.
        ##
        tries = 1
        while order_details["status"] != "closed" and tries <= 3:
            order_details = self.get_orders_info_with_retry(
                txid=txid,
                exit_on_fail=False,
            )
            LOG.warning(
                "Order '%s' is not closed! Retry %d/3 in %d seconds...",
                txid,
                tries,
                (wait_time := 2 + tries),
            )
            sleep(wait_time)
            tries += 1

        if order_details["status"] != "closed":
            LOG.warning(
                "Can not handle filled order, since the fetched order is not"
                " closed in upstream!"
                " This may happen due to Kraken's websocket API being faster"
                " than their REST backend. Retrying in a few seconds...",
            )
            self.handle_filled_order_event(txid=txid)
            return

        # ======================================================================
        if self.__s.dry_run:
            LOG.info("Dry run, not handling filled order event.")
            return

        # ======================================================================
        # Notify about the executed order
        ##
        self.__s.t.send_to_telegram(
            message=str(
                f"✅ {self.__s.symbol}: "
                f"{order_details['descr']['type'][0].upper()}{order_details['descr']['type'][1:]} "
                "order executed"
                f"\n ├ Price » {order_details['descr']['price']} {self.__s.quote_currency}"
                f"\n ├ Size » {order_details['vol_exec']} {self.__s.base_currency}"
                f"\n └ Size in {self.__s.quote_currency} » "
                f"{round(float(order_details['descr']['price']) * float(order_details['vol_exec']), self.__s.cost_decimals)}",
            ),
        )

        # ======================================================================
        # Create a sell order for the executed buy order.
        ##
        if order_details["descr"]["type"] == "buy":
            self.handle_arbitrage(
                side="sell",
                order_price=self.__s.get_order_price(
                    side="sell",
                    last_price=float(order_details["descr"]["price"]),
                ),
                txid_to_delete=txid,
            )

        # ======================================================================
        # Create a buy order for the executed sell order.
        ##
        elif (
            self.__s.orderbook.count(filters={"side": "sell"}, exclude={"txid": txid})
            != 0
        ):
            # A new buy order will only be placed if there is another sell
            # order, because if the last sell order was filled, the price is so
            # high, that all buy orders will be canceled anyway and new buy
            # orders will be placed in ``check_price_range`` during shift-up.
            self.handle_arbitrage(
                side="buy",
                order_price=self.__s.get_order_price(
                    side="buy",
                    last_price=float(order_details["descr"]["price"]),
                ),
                txid_to_delete=txid,
            )
        else:
            # Remove filled order from list of all orders
            self.__s.orderbook.remove(filters={"txid": txid})

    def handle_cancel_order(self: OrderManager, txid: str) -> None:
        """
        Cancels an order by txid, removes it from the orderbook, and checks if
        there there was some volume executed which can be sold later.

        NOTE: The orderbook is the "gate keeper" of this function. If the order
              is not present in the local orderbook, nothing will happen.

        For post-only buy orders - if these were cancelled by Kraken, they are
        still in the local orderbook and will be handled just like regular calls
        of the handle_cancel_order of the algorithm.

        For orders that were cancelled by the algorithm, these will cancelled
        via API and removed from the orderbook. The incoming "canceled" message
        by the websocket will be ignored, as the order is already removed from
        the orderbook.

        """
        if self.__s.orderbook.count(filters={"txid": txid}) == 0:
            return

        order_details = self.get_orders_info_with_retry(txid=txid)

        if (
            order_details["descr"]["pair"] != self.__s.altname
            or order_details["userref"] != self.__s.userref
        ):
            return

        if self.__s.dry_run:
            LOG.info("DRY RUN: Not cancelling order: %s", txid)
            return

        LOG.info("Cancelling order: '%s'", txid)

        try:
            self.__s.trade.cancel_order(txid=txid)
        except KrakenUnknownOrderError:
            LOG.info(
                "Order '%s' is already closed. Removing from orderbook...",
                txid,
            )

        self.__s.orderbook.remove(filters={"txid": txid})

        # Check if the order has some vol_exec to sell
        ##
        if float(order_details["vol_exec"]) != 0.0:
            LOG.info(
                "Order '%s' is partly filled - saving those funds.",
                txid,
            )
            b = self.__s.configuration.get()

            # Add vol_exec to remaining funds
            updates = {
                "vol_of_unfilled_remaining": b["vol_of_unfilled_remaining"]
                + float(order_details["vol_exec"]),
            }

            # Set new highest buy price.
            if b["vol_of_unfilled_remaining_max_price"] < float(
                order_details["descr"]["price"],
            ):
                updates |= {
                    "vol_of_unfilled_remaining_max_price": float(
                        order_details["descr"]["price"],
                    ),
                }
            self.__s.configuration.update(updates)

            # Sell remaining funds if there is enough to place a sell order.
            # Its not perfect but good enough. (Some funds may still be
            # stuck) - but better than nothing.
            b = self.__s.configuration.get()
            if (
                b["vol_of_unfilled_remaining"]
                * b["vol_of_unfilled_remaining_max_price"]
                >= self.__s.amount_per_grid
            ):
                LOG.info(
                    "Collected enough funds via partly filled buy orders to"
                    " create a new sell order...",
                )
                self.handle_arbitrage(
                    side="sell",
                    order_price=self.__s.get_order_price(
                        side="sell",
                        last_price=b["vol_of_unfilled_remaining_max_price"],
                    ),
                )
                self.__s.configuration.update(  # Reset the remaining funds
                    {
                        "vol_of_unfilled_remaining": 0,
                        "vol_of_unfilled_remaining_max_price": 0,
                    },
                )

    def cancel_all_open_buy_orders(self: OrderManager) -> None:
        """
        Cancels all open buy orders and removes them from the orderbook.
        """
        LOG.info("Cancelling all open buy orders...")
        for txid, order in self.__s.user.get_open_orders(
            userref=self.__s.userref,
        )["open"].items():
            if (
                order["descr"]["type"] == "buy"
                and order["descr"]["pair"] == self.__s.altname
            ):
                self.handle_cancel_order(txid=txid)
                sleep(0.2)  # Avoid rate limiting

        self.__s.orderbook.remove(filters={"side": "buy"})

    def get_orders_info_with_retry(
        self: OrderManager,
        txid: str,
        tries: int = 0,
        max_tries: int = 5,
        exit_on_fail: bool = True,
    ) -> dict | None:
        """
        Returns the order details for a given txid.

        NOTE: We need retry here, since Kraken lacks of fast processing of
              placed/filled orders and making them available via REST API.
        """
        while tries < max_tries and not (
            order_details := self.__s.user.get_orders_info(
                txid=txid,
            ).get(txid)
        ):
            tries += 1
            LOG.warning(
                "Could not find order '%s'. Retry %d/%d in %d seconds...",
                txid,
                tries,
                max_tries,
                (wait_time := 2 * tries),
            )
            sleep(wait_time)

        if exit_on_fail and order_details is None:
            self.__s.save_exit(
                f"Failed to retrieve order info for '{txid}' after"
                f" {max_tries} retries!",
            )

        order_details["txid"] = txid
        return order_details  # type: ignore[no-any-return]



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2023 Benjamin Thomas Schwertfeger
# All rights reserved.
# https://github.com/btschwertfeger
#

import sys
from logging import DEBUG, INFO, WARNING, basicConfig, getLogger
from typing import Any

from click import BOOL, FLOAT, INT, STRING, Context, echo, pass_context
from cloup import Choice, HelpFormatter, HelpTheme, Style, group, option, option_group
from cloup.constraints import If, accept_none, require_all


def print_version(ctx: Context, param: Any, value: Any) -> None:  # noqa: ANN401, ARG001
    """Prints the version of the package"""
    if not value or ctx.resilient_parsing:
        return
    from importlib.metadata import (  # noqa: PLC0415 # pylint: disable=import-outside-toplevel
        version,
    )

    echo(version("kraken-infinity-grid"))
    ctx.exit()


def ensure_larger_than_zero(
    ctx: Context,
    param: Any,  # noqa: ANN401
    value: Any,  # noqa: ANN401
) -> Any:  # noqa: ANN401
    """Ensure the value is larger than 0"""
    if value <= 0:
        ctx.fail(f"Value for option '{param.name}' must be larger than 0!")
    return value


def ensure_larger_equal_zero(
    ctx: Context,
    param: Any,  # noqa: ANN401
    value: Any,  # noqa: ANN401
) -> Any:  # noqa: ANN401
    """Ensure the value is larger than 0"""
    if value is not None and value < 0:
        ctx.fail(f"Value for option '{param.name}' must be larger then or equal to 0!")
    return value


@group(
    context_settings={
        "auto_envvar_prefix": "KRAKEN",
        "help_option_names": ["-h", "--help"],
    },
    formatter_settings=HelpFormatter.settings(
        theme=HelpTheme(
            invoked_command=Style(fg="bright_yellow"),
            heading=Style(fg="bright_white", bold=True),
            constraint=Style(fg="magenta"),
            col1=Style(fg="bright_yellow"),
        ),
    ),
    no_args_is_help=True,
)
@option(
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
)
@option(
    "--api-key",
    required=True,
    help="The Kraken Spot API key",
    type=STRING,
)
@option(
    "--secret-key",
    required=True,
    type=STRING,
    help="The Kraken Spot API secret key",
)
@option(
    "-v",
    "--verbose",
    count=True,
    help="Increase the verbosity of output. Use -vv for even more verbosity.",
)
@option(
    "--dry-run",
    required=False,
    is_flag=True,
    default=False,
    help="Enable dry-run mode which do not execute trades.",
)
@pass_context
def cli(ctx: Context, **kwargs: dict) -> None:
    """
    Command-line interface entry point
    """
    ctx.ensure_object(dict)
    ctx.obj |= kwargs

    verbosity = kwargs.get("verbose", 0)

    basicConfig(
        format="%(asctime)s %(levelname)8s | %(message)s",
        datefmt="%Y/%m/%d %H:%M:%S",
        level=INFO if verbosity == 0 else DEBUG,
    )

    getLogger("requests").setLevel(WARNING)
    getLogger("urllib3").setLevel(WARNING)
    getLogger("websockets").setLevel(WARNING)

    if verbosity > 1:  # type: ignore[operator]
        getLogger("requests").setLevel(DEBUG)
        getLogger("websockets").setLevel(DEBUG)
        getLogger("kraken").setLevel(DEBUG)
    else:
        getLogger("websockets").setLevel(WARNING)
        getLogger("kraken").setLevel(WARNING)


@cli.command(
    context_settings={
        "auto_envvar_prefix": "KRAKEN_RUN",
        "help_option_names": ["-h", "--help"],
    },
    formatter_settings=HelpFormatter.settings(
        theme=HelpTheme(
            invoked_command=Style(fg="bright_yellow"),
            heading=Style(fg="bright_white", bold=True),
            constraint=Style(fg="magenta"),
            col1=Style(fg="bright_yellow"),
        ),
    ),
)
@option(
    "--strategy",
    type=Choice(choices=("cDCA", "GridHODL", "GridSell", "SWING"), case_sensitive=True),
    help="The strategy to run.",
    required=True,
)
@option(
    "--name",
    required=True,
    type=STRING,
    help="""
    The name of the instance. Can be any name that is used to differentiate
    between instances of the kraken-infinity-grid.
    """,
)
@option(
    "--base-currency",
    required=True,
    type=STRING,
    help="The base currency.",
)
@option(
    "--quote-currency",
    required=True,
    type=STRING,
    help="The quote currency.",
)
@option(
    "--amount-per-grid",
    required=True,
    type=FLOAT,
    help="The quote amount to use per interval.",
)
@option(
    "--interval",
    required=True,
    type=FLOAT,
    default=0.04,
    callback=ensure_larger_than_zero,
    help="The interval between orders.",
)
@option(
    "--n-open-buy-orders",
    required=True,
    type=INT,
    default=3,
    callback=ensure_larger_than_zero,
    help="""
    The number of concurrent open buy orders e.g., ``5``. The number of
    always open buy positions specifies how many buy positions should be
    open at the same time. If the interval is defined to 2%, a number of 5
    open buy positions ensures that a rapid price drop of almost 10% that
    can be caught immediately.
    """,
)
@option(
    "--telegram-token",
    required=False,
    type=STRING,
    help="The Telegram token to use.",
)
@option(
    "--telegram-chat-id",
    required=False,
    type=STRING,
    help="The telegram chat ID to use.",
)
@option(
    "--exception-token",
    required=False,
    type=STRING,
    help="The telegram token to use for exceptions.",
)
@option(
    "--exception-chat-id",
    required=False,
    type=STRING,
    help="The telegram chat ID to use for exceptions.",
)
@option(
    "--max-investment",
    required=False,
    type=FLOAT,
    default=10e10,
    callback=ensure_larger_than_zero,
    help="""
    The maximum investment, e.g. 1000 USD that the algorithm will manage.
    """,
)
@option(
    "--userref",
    required=True,
    type=INT,
    callback=ensure_larger_than_zero,
    help="""
    A reference number to identify the algorithm's orders. This can be a
    timestamp or any integer number. Use different userref's for different
    instances!
    """,
)
@option(
    "--fee",
    type=FLOAT,
    required=False,
    callback=ensure_larger_equal_zero,
    help="""
    The fee percentage to respect, e.g. '0.0026' for 0.26 %. This value does not
    change the actual paid fee, instead it used to estimate order sizes. If not
    passed, the highest maker fee will be used.
    """,
)
@option(
    "--sqlite-file",
    type=STRING,
    help="SQLite file to use as database.",
)
@option(
    "--db-name",
    type=STRING,
    default="kraken_infinity_grid",
    help="The database name.",
)
@option_group(
    "PostgreSQL Database Options",
    option(
        "--db-user",
        type=STRING,
        help="PostgreSQL DB user",
    ),
    option(
        "--db-password",
        type=STRING,
        help="PostgreSQL DB password",
    ),
    option(
        "--db-host",
        type=STRING,
        help="PostgreSQL DB host",
    ),
    option(
        "--db-port",
        type=STRING,
        help="PostgreSQL DB port",
    ),
    constraint=If(
        "sqlite_file",
        then=accept_none,
        else_=require_all,
    ),
)
@pass_context
def run(ctx: Context, **kwargs: dict) -> None:
    """Run the trading algorithm using the specified options."""
    # pylint: disable=import-outside-top-level
    import asyncio  # noqa: PLC0415
    import traceback  # noqa: PLC0415

    from kraken_infinity_grid.gridbot import KrakenInfinityGridBot  # noqa: PLC0415

    db_config = {
        "sqlite_file": kwargs.pop("sqlite_file"),
        "db_user": kwargs.pop("db_user"),
        "db_password": kwargs.pop("db_password"),
        "db_host": kwargs.pop("db_host"),
        "db_port": kwargs.pop("db_port"),
        "db_name": kwargs.pop("db_name"),
    }
    ctx.obj |= kwargs

    async def main() -> None:
        # Instantiate the trading algorithm
        gridbot = KrakenInfinityGridBot(
            key=ctx.obj.pop("api_key"),
            secret=ctx.obj.pop("secret_key"),
            dry_run=ctx.obj.pop("dry_run"),
            config=kwargs,
            db_config=db_config,
        )

        try:
            await gridbot.run()
        except KeyboardInterrupt as exc:
            gridbot.save_exit(
                reason=f"Exception in top-run: {exc} {traceback.format_exc()}",
            )
            sys.exit(1)

    asyncio.run(main())


@cli.command(
    context_settings={
        "auto_envvar_prefix": "KRAKEN_CANCEL",
        "help_option_names": ["-h", "--help"],
    },
    formatter_settings=HelpFormatter.settings(
        theme=HelpTheme(
            invoked_command=Style(fg="bright_yellow"),
            heading=Style(fg="bright_white", bold=True),
            constraint=Style(fg="magenta"),
            col1=Style(fg="bright_yellow"),
        ),
    ),
)
@option(
    "-f",
    "--force",
    required=False,
    type=BOOL,
    default=False,
    is_flag=True,
    show_default=True,
)
@pass_context
def cancel(ctx: Context, **kwargs: dict) -> None:
    """Cancel all open orders."""
    ctx.obj |= kwargs
    if not ctx.obj["force"]:
        print("Not canceling unless '-f' was passed!")  # noqa: T201
        sys.exit(1)

    from pprint import pprint  # noqa: PLC0415

    from kraken.spot import Trade  # noqa: PLC0415

    trade = Trade(key=ctx.obj["api_key"], secret=ctx.obj["secret_key"])
    pprint(trade.cancel_all_orders())  # noqa: T203
